/* wrap.m     Simple custom wrappers for IcosaBlue example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 */
# import <stdlib.h>
# import <objc/runtime.h>
# import <Quartzcore/Quartzcore.h>
# import <UIKit/UIKit.h>
# import <OpenGLES/EAGLDrawable.h>
# import <caml/mlvalues.h>
# import <caml/memory.h>
# import <caml/alloc.h>
# import <caml/callback.h>
# import "ViewDelegator.h"
# import "wrap.h"

/* Represent raw ObjC object pointers in OCaml as nativeint values.
 *
 * Note: to keep things simple, we retain ObjC values and don't release
 * them.  It works for this example because the few ObjC objects
 * accessible from the OCaml side are never destroyed.
 * 
 * In a more dynamic setting, need to gain control when the OCaml
 * wrapper objects are GCed and release the wrapped up ObjC objects.
 */
static value wrapRawObjC(id obj) {
    CAMLparam0();
    [obj retain];
    CAMLreturn(caml_copy_nativeint((intnat) obj));
}

static id unwrapRawObjC(value objcval) {
    return (id) Nativeint_val(objcval);
}

/* Need to be able to determine the OCaml object that wraps a given ObjC
 * object.  In the simple setup here, the association is permanent and
 * there are very few objects involved.  So we can use a little table.
 * (Getting the ObjC object that wraps an OCaml object is easier; we
 * just use the container method of Wrappee.t.)
 */
static struct {
    id objc_obj;
    value ocaml_obj;
} g_ocaml_wrapper[16];
static int g_ocwrap_ct = 0;


static value find_ocaml_wrapper(id obj) {
    int i;

    for(i = 0; i < g_ocwrap_ct; i++)
        if(g_ocaml_wrapper[i].objc_obj == obj)
            return g_ocaml_wrapper[i].ocaml_obj;
    return 0;
}


static value wrapObjC(char *wrapcs, id obj) {
    /* Create a new OCaml object that wraps the given ObjC object.
     * wrapcs gives the name of the closure for creating an OCaml
     * wrapper of the correct type.  A closure is registered by each
     * OCaml class that wraps an ObjC class (i.e., each subclass of
     * Wrapper.t).
     */
    CAMLparam0();
    CAMLlocal1(val);
    value *closure;

    closure = (value *) caml_named_value(wrapcs);
    if(closure == NULL) {
        /* Can't create an instance of an unregistered class.  If this
         * happens, things will go very bad from here.
         */
        printf("wrapObjC lookup failure for %s\n", wrapcs);
        fflush(stdout);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    val = caml_callback(*closure, wrapRawObjC(obj));

    /* Enter the wrapping into the table.
     */
    if(g_ocwrap_ct >= sizeof g_ocaml_wrapper / sizeof g_ocaml_wrapper[0]) {
        /* Things might go bad from here if we need to access the
         * wrapper later from the ObjC side.  For this simple example it
         * never happens because there are only 2 or 3 wrapped ObjC
         * objects.
         */
        printf("wrapObjC table overflow for %s\n", wrapcs);
        fflush(stdout);
    } else {
        g_ocaml_wrapper[g_ocwrap_ct].objc_obj = obj;
        g_ocaml_wrapper[g_ocwrap_ct].ocaml_obj = val;
        caml_register_global_root(&g_ocaml_wrapper[g_ocwrap_ct].ocaml_obj);
        g_ocwrap_ct++;
    }

    CAMLreturn(val);
}

static id unwrapObjC(value obj) {
    /* Return the ObjC object wrapped by the given OCaml object.
     */
    CAMLparam1(obj);
    CAMLlocal1(objcval);

    objcval =
        caml_callback(
            caml_get_public_method(obj, caml_hash_variant("contents")),
            obj);
    CAMLreturnT(id, unwrapRawObjC(objcval));
}


/* Subclasses of WrapOCaml wrap OCaml objects as ObjC objects.
 */
@implementation WrapOCaml {
}

@dynamic contents;

- (value) contents
{
    return contents;
}


- (void) setContents: (value) aValue
{
    contents = aValue;
}


- (WrapOCaml *) init
{
    /* Create a new OCaml object that is wrapped by self.  We match the
     * OCaml class to our own class by name--simple but effective.  We
     * then call the closure for creating a wrapped object, which is
     * registered by each OCaml class that can be wrapped by an ObjC
     * class (i.e., each subtype of Wrappee.t).
     */
#   define CLOSURE_SFX      ".wrapped"
#   define CLOSURE_SFX_LEN  (sizeof CLOSURE_SFX - 1)
    const char *classname;
    char closurename[128];
    value *closure;

    if((self = [super init]) != nil) {
        classname = object_getClassName(self);
        if(strlen(classname) + CLOSURE_SFX_LEN >= sizeof closurename) {
            printf("[WrapOCaml init]: class name too long: %s\n", classname);
            fflush(stdout);
            return nil;
        }
        strcpy(closurename, classname);
        strcat(closurename, CLOSURE_SFX);
        closure = (value *) caml_named_value(closurename);
        if(closure == NULL) {
            printf("[WrapOCaml init]: lookup failure for %s\n", closurename);
            fflush(stdout);
            return nil;
        }
        contents = caml_callback(*closure, wrapRawObjC(self));
        caml_register_global_root(&contents);
    }
    return self;
}

- (void) dealloc
{
    if(contents != 0) {
        caml_remove_global_root(&contents);
        contents = 0;
    }
    [super dealloc];
}

@end

/* Conversion utilities.
 */
static value Val_fourtuple(double a, double b, double c, double d)
{
    CAMLparam0();
    CAMLlocal1(res);

    res = caml_alloc_tuple(4);
    Store_field(res, 0, caml_copy_double(a));
    Store_field(res, 1, caml_copy_double(b));
    Store_field(res, 2, caml_copy_double(c));
    Store_field(res, 3, caml_copy_double(d));

    CAMLreturn(res);
}

static void Fourtuple_val(double *ap, double *bp, double *cp, double *dp,
                        value tuple)
{
    CAMLparam1(tuple);

    *ap = Double_val(Field(tuple, 0));
    *bp = Double_val(Field(tuple, 1));
    *cp = Double_val(Field(tuple, 2));
    *dp = Double_val(Field(tuple, 3));

    CAMLreturn0;
}

static value Val_point(CGPoint *point)
{
    CAMLparam0();
    CAMLlocal1(res);

    res = caml_alloc_tuple(2);
    Store_field(res, 0, caml_copy_double(point->x));
    Store_field(res, 1, caml_copy_double(point->y));

    CAMLreturn(res);
}

static void Point_val(CGPoint *ptp, value pointval)
{
    CAMLparam1(pointval);

    ptp->x = Double_val(Field(pointval, 0));
    ptp->y = Double_val(Field(pointval, 1));

    CAMLreturn0;
}

static value Val_rect(CGRect *rectp)
{
    return
        Val_fourtuple(rectp->origin.x, rectp->origin.y,
                        rectp->size.width, rectp->size.height);
}

static void Rect_val(CGRect *rectp, value rectval)
{
    double res[4];

    Fourtuple_val(res + 0, res + 1, res + 2, res + 3, rectval);
    rectp->origin.x = res[0];
    rectp->origin.y = res[1];
    rectp->size.width = res[2];
    rectp->size.height = res[3];
}

static SEL SEL_val(value selectval)
{
    /* Translate the OCaml string into an ObjC selector.
     *
     * Conventionally we use '\'' in OCaml where ':' appears in ObjC.
     * So here we translate back.
     */
    char *prime;
    SEL res;

    char *ocamlsel = String_val(selectval);
    char *buf = malloc(strlen(ocamlsel) + 1);
    strcpy(buf, ocamlsel);
    prime = buf;
    while((prime = strchr(prime, '\'')) != NULL)
        *prime = ':';
    res = sel_registerName(buf);
    free(buf);
    return res;
}


/* OCaml objects accessed from ObjC.
 */

@implementation Icosactlr : WrapOCaml
                    /* <UIApplicationDelegate, ViewDelegate> */
{
}

@dynamic delegator;

- (UIView *) delegator
{
    CAMLparam0();
    CAMLlocal2(selfval, delegatorval);

    selfval = [self contents];
    delegatorval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("delegator")),
            selfval);
    CAMLreturnT(UIView *, unwrapObjC(delegatorval));
}

- (void) setDelegator: (UIView *) aView
{
    CAMLparam0();
    CAMLlocal2(selfval, delegatorval);

    selfval = [self contents];
    delegatorval = wrapObjC("UIView.wrap", (id) aView);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setDelegator'")),
            selfval,
            delegatorval);
    CAMLreturn0;
}

- (void) applicationDidFinishLaunching: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidFinishLaunching'")),
            selfval,
            applval);
    CAMLreturn0;
}

#ifdef NOT_NEEDED
- (void) applicationDidBecomeActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidBecomeActive'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationWillResignActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationWillResignActive'")),
            selfval,
            applval);
    CAMLreturn0;
}
#endif /* NOT_NEEDED */

- (void) applicationDidEnterBackground: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidEnterBackground'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationWillEnterForeground: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationWillEnterForeground'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationDidReceiveMemoryWarning: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidReceiveMemoryWarning'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) viewLayoutSubviews: (UIView *) aView
{
    CAMLparam0();
    CAMLlocal2(selfval, viewval);

    selfval = [self contents];

    if((viewval = find_ocaml_wrapper(aView)) == 0)
        viewval = wrapObjC("UIView.wrap", (id) aView);

    (void)
        caml_callback2(
            caml_get_public_method(selfval,
                                   caml_hash_variant("viewLayoutSubviews'")),
            selfval,
            viewval);
    CAMLreturn0;
}

- (void) drawView: (NSTimer *) aTimer
{
    CAMLparam0();
    CAMLlocal2(selfval, timerval);

    selfval = [self contents];

    if((timerval = find_ocaml_wrapper(aTimer)) == 0)
        timerval = wrapObjC("NSTimer.wrap", (id) aTimer);

    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("drawView'")),
            selfval,
            timerval);
    CAMLreturn0;
}

@end


/* ObjC objects accessed from OCaml.
 */

/* NSTimer
 */

value NSTimer_scheduledTimer5(value timeInterval, value target,
                    value selectval, value userInfo, value repeats)
    /* For simplicity, the target must be wrapped (by an ObjC object)
     * already.  It's OK if userInfo isn't wrapped; we treat this as
     * nil.
     */
{
    CAMLparam5(timeInterval, target, selectval, userInfo, repeats);
    CAMLlocal2(targro, uiro);

    double ti = Double_val(timeInterval);
    targro = 
        caml_callback(
            caml_get_public_method(target, caml_hash_variant("container")),
            target);
    id targ = unwrapRawObjC(targro);
    if(targ == nil) {
        /* If the target wasn't already wrapped, things will go very bad
         * from here.
         */
        printf("NSTimer_scheduledTimer5: unwrapped target\n");
        fflush(stdout);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    SEL sel = SEL_val(selectval);
    uiro =
        caml_callback(
            caml_get_public_method(userInfo, caml_hash_variant("container")),
            userInfo);
    id ui = unwrapRawObjC(uiro);
    int rep = Bool_val(repeats);

    NSTimer *timer =
        [NSTimer scheduledTimerWithTimeInterval: ti
                 target: targ
                 selector: sel
                 userInfo: ui
                 repeats: rep ];
    CAMLreturn(wrapObjC("NSTimer.wrap", (id) timer));
}

/* CGContext
 */

/* (These are copied from ml_raw.h and raw_tags.h in the LablGLES
 * implementation.  It makes it easier to build the test.  For
 * production work, #include instead of copying!)
 */
#define Base_raw(raw) (Field(raw,1))
#define Offset_raw(raw) (Field(raw,2))
#define Size_raw(raw) (Field(raw,3))
#define Addr_raw(raw) (Base_raw(raw)+Long_val(Offset_raw(raw)))
#define Byte_raw(raw) ((char *) Addr_raw(raw))

#define MLTAG_ubyte     Val_int(520420861)

extern value ml_raw_alloc_static(value kind, value len);


value CGContext_create(value widthval, value heightval)
/* int -> int -> [`ubyte] Raw.t * nativeint */
{
    /* The underlying raw value must be statically allocated because the
     * Quartz bitmap context could never cope with the changing
     * addresses produced by normal OCaml memory management.  Currently
     * there's no way to deallocate this value, which is fine for a
     * small test program.  In production you'd have to get control at
     * the deallocation of the CgContext.t value and free the raw value
     * at that time.
     */
    CAMLparam2(widthval, heightval);
    CAMLlocal2(rawval, resval);
    CGColorSpaceRef csp;
    CGContextRef res;
    int width = Int_val(widthval);
    int height = Int_val(heightval);

    if(width < 0 || height < 0)
        caml_invalid_argument("CgContext.create size");
    rawval = ml_raw_alloc_static(MLTAG_ubyte, Val_int(width * height * 4));
    csp = CGColorSpaceCreateDeviceRGB();
    res =
        CGBitmapContextCreate(Byte_raw(rawval), width, height, 8, width * 4,
            csp, kCGImageAlphaPremultipliedLast);
    CGContextSetFillColorSpace(res, csp);
    CGContextSetStrokeColorSpace(res, csp);
    CGColorSpaceRelease(csp);
    resval = caml_alloc(2, 0);
    Store_field(resval, 0, rawval);
    Store_field(resval, 1, caml_copy_nativeint((intnat) res));
    CAMLreturn(resval);
}

value CGContext_setLineCap(value ctxval, value capval)
/* [`ubyte] Raw.t * nativeint -> linecap -> unit */
{
    CAMLparam2(ctxval, capval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextSetLineCap(ctx, Int_val(capval)); /* capval translates directly */
    CAMLreturn(Val_unit);
}

value CGContext_setLineWidth(value ctxval, value widthval)
/* [`ubyte] Raw.t * nativeint -> float -> unit */
{
    CAMLparam2(ctxval, widthval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextSetLineWidth(ctx, Double_val(widthval));
    CAMLreturn(Val_unit);
}

value CGContext_addCurveToPoint(value ctxval, value cp1val, value cp2val,
                    value pval)
/* [`ubyte] Raw.t * nativeint ->
           Cocoa.point -> Cocoa.point -> Cocoa.point -> unit */
{
    CAMLparam4(ctxval, cp1val, cp2val, pval);
    CGContextRef ctx;
    CGPoint cp1, cp2, p;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    Point_val(&cp1, cp1val);
    Point_val(&cp2, cp2val);
    Point_val(&p, pval);

    CGContextAddCurveToPoint(ctx, cp1.x, cp1.y, cp2.x, cp2.y, p.x, p.y);

    CAMLreturn(Val_unit);
}

value CGContext_addLineToPoint(value ctxval, value pval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.point -> unit */
{
    CAMLparam2(ctxval, pval);
    CGContextRef ctx;
    CGPoint p;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    Point_val(&p, pval);

    CGContextAddLineToPoint(ctx, p.x, p.y);

    CAMLreturn(Val_unit);
}

value CGContext_beginPath(value ctxval)
/* [`ubyte] Raw.t * nativeint -> unit */
{
    CAMLparam1(ctxval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextBeginPath(ctx);
    CAMLreturn(Val_unit);
}

value CGContext_closePath(value ctxval)
/* [`ubyte] Raw.t * nativeint -> unit */
{
    CAMLparam1(ctxval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextClosePath(ctx);
    CAMLreturn(Val_unit);
}

value CGContext_moveToPoint(value ctxval, value pval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.point -> unit */
{
    CAMLparam2(ctxval, pval);
    CGContextRef ctx;
    CGPoint p;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    Point_val(&p, pval);

    CGContextMoveToPoint(ctx, p.x, p.y);

    CAMLreturn(Val_unit);
}

value CGContext_fillPath(value ctxval)
/* [`ubyte] Raw.t * nativeint -> unit */
{
    CAMLparam1(ctxval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextFillPath(ctx);
    CAMLreturn(Val_unit);
}

value CGContext_fillRect(value ctxval, value rectval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.rect -> unit */
{
    CAMLparam2(ctxval, rectval);
    CGContextRef ctx;
    CGRect r;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    Rect_val(&r, rectval);

    CGContextFillRect(ctx, r);

    CAMLreturn(Val_unit);
}

value CGContext_strokePath(value ctxval)
/* [`ubyte] Raw.t * nativeint -> unit */
{
    CAMLparam1(ctxval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextStrokePath(ctx);
    CAMLreturn(Val_unit);
}

value CGContext_setFillColor(value ctxval, value colorval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.color -> unit */
{
    CAMLparam2(ctxval, colorval);
    CGContextRef ctx;
    CGFloat components[4];

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    components[0] = Double_val(Field(colorval, 0));
    components[1] = Double_val(Field(colorval, 1));
    components[2] = Double_val(Field(colorval, 2));
    components[3] = Double_val(Field(colorval, 3));

    CGContextSetFillColor(ctx, components);

    CAMLreturn(Val_unit);
}

value CGContext_setStrokeColor(value ctxval, value colorval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.color -> unit */
{
    CAMLparam2(ctxval, colorval);
    CGContextRef ctx;
    CGFloat components[4];

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    components[0] = Double_val(Field(colorval, 0));
    components[1] = Double_val(Field(colorval, 1));
    components[2] = Double_val(Field(colorval, 2));
    components[3] = Double_val(Field(colorval, 3));

    CGContextSetStrokeColor(ctx, components);

    CAMLreturn(Val_unit);
}

value CGContext_scaleCTM(value ctxval, value mxval, value myval)
/* [`ubyte] Raw.t * nativeint -> float -> float -> unit */
{
    CAMLparam3(ctxval, mxval, myval);
    CGContextRef ctx;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));

    CGContextScaleCTM(ctx, Double_val(mxval), Double_val(myval));

    CAMLreturn(Val_unit);
}

value CGContext_translateCTM(value ctxval, value dxval, value dyval)
/* [`ubyte] Raw.t * nativeint -> float -> float -> unit */
{
    CAMLparam3(ctxval, dxval, dyval);
    CGContextRef ctx;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));

    CGContextTranslateCTM(ctx, Double_val(dxval), Double_val(dyval));

    CAMLreturn(Val_unit);
}

value CGContext_getTextPosition(value ctxval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.point */
{
    CAMLparam1(ctxval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGPoint p = CGContextGetTextPosition(ctx);
    CAMLreturn(Val_point(&p));
}

value CGContext_selectFont(value ctxval, value fstrval, value sizeval)
/* [`ubyte] Raw.t * nativeint -> string -> float -> unit */
{
    /* For this simple test program, just support MacRoman encoding.
     */
    CAMLparam3(ctxval, fstrval, sizeval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextSelectFont(ctx, String_val(fstrval), Double_val(sizeval),
                        kCGEncodingMacRoman);
    CAMLreturn(Val_unit);
}

value CGContext_setTextDrawingMode(value ctxval, value tdval)
/* [`ubyte] Raw.t * nativeint -> tdmode -> unit */
{
    CAMLparam2(ctxval, tdval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextSetTextDrawingMode(ctx, Int_val(tdval)); /* Translates directly */
    CAMLreturn(Val_unit);
}

value CGContext_setTextMatrix(value ctxval, value atmval)
/* [`ubyte] Raw.t * nativeint -> atm -> unit */
{
    CAMLparam2(ctxval, atmval);
    CGContextRef ctx;
    CGAffineTransform atm;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    atm.a = Double_val(Field(atmval, 0));
    atm.b = Double_val(Field(atmval, 1));
    atm.c = Double_val(Field(atmval, 2));
    atm.d = Double_val(Field(atmval, 3));
    atm.tx = Double_val(Field(atmval, 4));
    atm.ty = Double_val(Field(atmval, 5));

    CGContextSetTextMatrix(ctx, atm);

    CAMLreturn(Val_unit);
}

value CGContext_setTextPosition(value ctxval, value ptval)
/* [`ubyte] Raw.t * nativeint -> Cocoa.point -> unit */
{
    CAMLparam2(ctxval, ptval);
    CGContextRef ctx;
    CGPoint p;

    ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    Point_val(&p, ptval);

    CGContextSetTextPosition(ctx, p.x, p.y);

    CAMLreturn(Val_unit);
}

value CGContext_showText(value ctxval, value strval, value lenval)
/* [`ubyte] Raw.t * nativeint -> string -> int -> unit */
{
    CAMLparam3(ctxval, strval, lenval);
    CGContextRef ctx = (CGContextRef) Nativeint_val(Field(ctxval, 1));
    CGContextShowText(ctx, String_val(strval), Int_val(lenval));
    CAMLreturn(Val_unit);
}

/* EAGLContext
 */
value EAGLContext_initWithAPI_(value objcval, value apival)
/* nativeint -> int -> EAGLContext.t */
{
    CAMLparam2(objcval, apival);
    EAGLContext *ctx, *ctxi;

    ctx = unwrapRawObjC(objcval);
    if(ctx == nil)
        ctx = [EAGLContext alloc];
    ctxi = [ctx initWithAPI: Int_val(apival)];
    [ctxi autorelease]; /* wrapObjC will retain */

    CAMLreturn(wrapObjC("EAGLContext.wrap", (id) ctxi));
}

value EAGLContext_renderbufferStorage_fromDrawable_(value objcval,
                    value targval, value drawbval)
/* nativeint -> int -> nativeint -> bool */
{
    CAMLparam3(objcval, targval, drawbval);

    EAGLContext *ctx;
    id <EAGLDrawable> drawb;
    BOOL res;

    ctx = unwrapRawObjC(objcval);
    drawb = unwrapRawObjC(drawbval);
    res = [ctx renderbufferStorage: Int_val(targval) fromDrawable: drawb];

    CAMLreturn(Val_bool(res));
}

value EAGLContext_presentRenderbuffer_(value objcval, value targval)
/* nativeint -> int -> bool */
{
    CAMLparam2(objcval, targval);
    BOOL res;
    EAGLContext *ctx = unwrapRawObjC(objcval);
    res = [ctx presentRenderbuffer: Int_val(targval)];
    CAMLreturn(Val_bool(res));
}

value EAGLContext_setCurrentContext_(value objcval)
/* nativeint -> bool */
{
    CAMLparam1(objcval);
    BOOL res;
    EAGLContext *ctx = unwrapRawObjC(objcval);
    res = [EAGLContext setCurrentContext: ctx];
    CAMLreturn(Val_bool(res));
}

/* CALayer
 */

value CALayer_isOpaque(value objcval)
/* nativeint -> bool */
{
    CAMLparam1(objcval);
    BOOL res;
    CALayer *layer = unwrapRawObjC(objcval);
    res = [layer isOpaque];
    CAMLreturn(Val_bool(res));
}

value CALayer_setOpaque_(value objcval, value boolval)
/* nativeint -> bool -> unit */
{
    CAMLparam2(objcval, boolval);
    CALayer *layer = unwrapRawObjC(objcval);
    [layer setOpaque: Bool_val(boolval)];
    CAMLreturn(Val_unit);
}


/* UIView
 */

value UIView_layer(value objcval)
/* nativeint -> CaLayer.t */
{
    CAMLparam1(objcval);
    UIView *view = unwrapRawObjC(objcval);
    CALayer *layer = [view layer];
    CAMLreturn(wrapObjC("CALayer.wrap", (id) layer));
}

value UIView_frame(value objcval)
/* nativeint -> Cocoa.rect */
{
    CAMLparam1(objcval);
    CGRect rect;

    UIView *view = unwrapRawObjC(objcval);
    rect = [view frame];
    CAMLreturn(Val_rect(&rect));
}

value UIView_setNeedsDisplay(value objcval)
/* nativeint -> unit */
{
    CAMLparam1(objcval);

    UIView *view = unwrapRawObjC(objcval);
    [view setNeedsDisplay];
    CAMLreturn(Val_unit);
}

/* UIApplication
 */

value UIApplication_statusBarStyle(value objcval)
/* nativeint -> sbstyle */
{
    CAMLparam1(objcval);
    UIStatusBarStyle sbstyle;

    UIApplication *appl = unwrapRawObjC(objcval);
    sbstyle = [appl statusBarStyle];

    CAMLreturn(Val_int(sbstyle)); /* sbstyle converts directly */
}

value UIApplication_setStatusBarStyle_(value objcval, value sbstyleval)
/* nativeint -> sbstyle -> unit */
{
    CAMLparam2(objcval, sbstyleval);

    UIApplication *appl = unwrapRawObjC(objcval);
    [appl setStatusBarStyle: Int_val(sbstyleval)]; /* converts directly */

    CAMLreturn(Val_unit);
}

value UIApplication_sharedApplication(value unitval)
/* unit -> nativeint */
{
    CAMLparam1(unitval);
    UIApplication *appl = [UIApplication sharedApplication];
    CAMLreturn(wrapObjC("UIApplication.wrap", appl));
}
